import django.db.models.deletion
from django.db import migrations, models


def move_episode_timeslot_relationship_to_timeslot(apps, schema_editor):
    TimeSlot = apps.get_model("program", "TimeSlot")
    for timeslot in TimeSlot.objects.filter(one_to_one_episode__isnull=False).select_related(
        "one_to_one_episode"
    ):
        timeslot.episode = timeslot.one_to_one_episode
        timeslot.save()


def move_timeslot_episode_relationship_to_episode(apps, schema_editor):
    Episode = apps.get_model("program", "Episode")
    for episode in Episode.objects.all():
        timeslots = list(episode.timeslots.all())
        timeslot_count_for_episode = len(timeslots)

        if timeslot_count_for_episode == 0:
            # The data model does not allow episodes to exist on their own, because
            # even though episode.timeslot is nullable it is also unique.
            episode.delete()
            continue

        if timeslot_count_for_episode == 1:
            # An episode with a single timeslot is easy
            # because we can simply assign it to the one-to-one relationship.
            episode.timeslot = timeslots[0]
            episode.save()
            continue

        # Multiple timeslots referenced the same episode.
        # We need to create copies of the episode for each timeslot.
        for timeslot in timeslots:
            new_episode = Episode.objects.create(
                timeslot=timeslot,
                created_at=episode.created_at,
                created_by=episode.created_by,
                updated_at=episode.updated_at,
                updated_by=episode.updated_by,
                title=episode.title,
                summary=episode.summary,
                content=episode.content,
                cba_id=episode.cba_id,
                image_id=episode.image_id,
                tags=episode.tags,
                memo=episode.memo,
                media_id=episode.media_id,
                show_id=episode.show_id,
            )
            new_episode.contributors.set(episode.contributors.all())
            new_episode.languages.set(episode.languages.all())
            new_episode.topics.set(episode.topics.all())
        # The episode can be safely deleted because it was duplicated for each timeslot.
        episode.delete()

    # TimeSlots do not exist without episodes in the old model so we need to ensure
    # that every timeslot has an episode assigned even if it is empty.
    TimeSlot = apps.get_model("program", "TimeSlot")
    timeslots_without_episodes = TimeSlot.objects.filter(one_to_one_episode__isnull=True)
    for timeslot_id, show_id in timeslots_without_episodes.values_list("id", "schedule__show_id"):
        Episode.objects.create(timeslot_id=timeslot_id, show_id=show_id)


class Migration(migrations.Migration):
    dependencies = [
        ("program", "0129_add_episode_show_relation"),
    ]

    operations = [
        # Reverse the relationship of episode and models.
        # First change the related name of the Episode.timeslot field,
        # so we can distinguish it from the newly introduced Timeslot.episode field.
        migrations.AlterField(
            model_name="episode",
            name="timeslot",
            field=models.OneToOneField(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="one_to_one_episode",
                to="program.timeslot",
            ),
        ),
        migrations.AddField(
            model_name="timeslot",
            name="episode",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="timeslots",
                to="program.episode",
            ),
        ),
        migrations.RunPython(
            move_episode_timeslot_relationship_to_timeslot,
            move_timeslot_episode_relationship_to_episode,
        ),
        migrations.RemoveField(
            model_name="episode",
            name="timeslot",
        ),
    ]
