import datetime

import pytest

from django.core.management import call_command
from django.utils import timezone

pytestmark = pytest.mark.django_db


# adapted from https://adamj.eu/tech/2024/06/23/django-test-pending-migrations/ for pytest
def test_no_pending_migrations(capsys):
    try:
        call_command("makemigrations", "--dry-run", "--check")
    except SystemExit:  # pragma: no cover
        out, _ = capsys.readouterr()

        raise AssertionError("Pending migrations:\n" + out) from None


@pytest.mark.django_db
@pytest.mark.migration
def test_note_to_episodes_migration(migrator):
    now = timezone.now()
    memo = "my memo"

    # create test state
    old_state = migrator.apply_initial_migration(
        ("program", "0127_alter_playlistentry_options_and_more")
    )
    Show = old_state.apps.get_model("program", "Show")
    Note = old_state.apps.get_model("program", "Note")
    TimeSlot = old_state.apps.get_model("program", "TimeSlot")
    Schedule = old_state.apps.get_model("program", "Schedule")
    Playlist = old_state.apps.get_model("program", "Playlist")
    show_a = Show.objects.create(name="show a", slug="show-a")
    schedule_show_a = Schedule.objects.create(
        show=show_a,
        first_date=now.date(),
        start_time=now.time(),
        end_time=(now + datetime.timedelta(hours=1)).time(),
    )
    playlist_show_a = Playlist.objects.create(show=show_a)
    timeslot_show_a = TimeSlot.objects.create(
        schedule=schedule_show_a,
        start=now,
        end=now + datetime.timedelta(hours=1),
        memo=memo,
        playlist=playlist_show_a,
    )
    note_show_a = Note.objects.create(
        timeslot=timeslot_show_a,
    )
    show_b = Show.objects.create(name="show b", slug="show-b")
    schedule_show_b = Schedule.objects.create(
        show=show_b,
        first_date=now.date(),
        start_time=now.time(),
        end_time=(now + datetime.timedelta(hours=1)).time(),
    )
    playlist_show_b = Playlist.objects.create(show=show_b)
    timeslot_show_b = TimeSlot.objects.create(
        schedule=schedule_show_b,
        start=now,
        end=now + datetime.timedelta(hours=1),
        memo=memo,
        playlist=playlist_show_b,
    )
    note_show_b = Note.objects.create(
        timeslot=timeslot_show_b,
    )

    # migrate to new state to test forward migration
    new_state = migrator.apply_tested_migration(("program", "0132_cleanup_episodes"))
    TimeSlot = new_state.apps.get_model("program", "TimeSlot")
    Episode = new_state.apps.get_model("program", "Episode")

    # The test data should be mapped 1:1.
    assert TimeSlot.objects.count() == 2
    assert Episode.objects.count() == 2
    # Check if the note is the same episode and has the same timeslot assigned.
    episode = Episode.objects.get(id=note_show_a.id)
    assert episode.id == note_show_a.id
    assert episode.timeslots.count() == 1
    assert episode.timeslots.first().id == timeslot_show_a.id
    # Check if the data from the timeslot was copied successfully.
    assert episode.media_id == playlist_show_a.id
    assert episode.memo == memo
    # Check if we assigned the correct show id from the schedule.
    assert episode.show_id == show_a.id
    assert Episode.objects.get(id=note_show_b.id).show_id == show_b.id


@pytest.mark.django_db
@pytest.mark.migration
def test_episode_to_note_migration(migrator):
    now = timezone.now()

    # create test state
    old_state = migrator.apply_initial_migration(("program", "0132_cleanup_episodes"))
    Show = old_state.apps.get_model("program", "Show")
    Episode = old_state.apps.get_model("program", "Episode")
    TimeSlot = old_state.apps.get_model("program", "TimeSlot")
    Schedule = old_state.apps.get_model("program", "Schedule")
    Playlist = old_state.apps.get_model("program", "Playlist")
    show = Show.objects.create(name="my test show", slug="my-test-show")
    Show.objects.create(name="another show", slug="another-show")
    schedule = Schedule.objects.create(
        show=show,
        first_date=now.date(),
        start_time=now.time(),
        end_time=(now + datetime.timedelta(hours=1)).time(),
    )
    playlist = Playlist.objects.create(show=show)
    episode = Episode.objects.create(
        title="my title",
        memo="my memo",
        media=playlist,
        show=show,
    )
    timeslot_to_episode_map = {}
    for offset, episode in [[0, episode], [7, episode], [14, None]]:
        timeslot = TimeSlot.objects.create(
            schedule=schedule,
            start=now + datetime.timedelta(days=offset),
            end=now + datetime.timedelta(days=offset, hours=1),
            episode=episode,
        )
        timeslot_to_episode_map[timeslot.id] = episode

    # migrate to old state to test reverse migration
    new_state = migrator.apply_tested_migration(
        ("program", "0127_alter_playlistentry_options_and_more")
    )
    TimeSlot = new_state.apps.get_model("program", "TimeSlot")
    Note = new_state.apps.get_model("program", "Note")

    # We should still have three timeslots.
    assert TimeSlot.objects.count() == 3
    # We should have an equal amount of notes.
    assert Note.objects.count() == 3

    # Check if episode data was successfully copied to the timeslot.
    for note in Note.objects.all():
        assert note.timeslot is not None
        episode = timeslot_to_episode_map[note.timeslot.id]
        if episode is None:
            # Timeslots that didn’t have an assigned episode should now be empty.
            assert note.timeslot.memo == ""
            assert note.timeslot.playlist_id is None
            # The note should be empty as well:
            assert note.title == ""
        else:
            # Timeslots with an episode assignment should now contain
            # the episode data.
            assert note.timeslot.memo == episode.memo
            assert note.timeslot.playlist_id == episode.media_id
            # The note should have been copied from the episode.
            assert note.title == episode.title
