from collections.abc import Iterable

from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied

from program.serializers import (
    SimpleAuditUpdaterMixin,
    UpdateFieldPermissionValidatorMixin,
)


class DummyUser:
    def __init__(self, permissions: Iterable[str] | None = None):
        self.username = "dummy user"
        self.permissions = set(permissions or [])

    def get_all_permissions(self):
        return set(self.permissions)

    def has_perm(self, perm, obj=None):
        return perm in self.permissions

    def has_perms(self, perms, obj=None):
        return all(perm in self.permissions for perm in perms)

    @classmethod
    def make_context(cls, permissions: Iterable[str] | None = None):
        class Request:
            user = cls(permissions=permissions)

        return {"request": Request()}


# It may be worthwhile to extract these tests to a steering-independent
# library so we don’t make user of any steering models here.
class _Episode:
    def __init__(self):
        self.title = "my title"
        self.subtitle = "my subtitle"
        self.content = "my content"
        self.created_by = ""
        self.updated_by = ""

    class _meta:
        app_label = "program"
        model_name = "episode"
        permissions = [
            ("edit__episode__title", "Can edit title field"),
            ("edit__episode__image", "Can edit image field"),
            ("edit__episode__contributors", "Can edit contributors field"),
        ]

    def save(self, *args, **kwargs):
        pass


class SimpleAssignmentWriter:
    def update(self, instance, validated_data):
        for key, value in validated_data.items():
            setattr(instance, key, value)
        return instance

    def create(self, validated_data):
        return self.update(self.Meta.model(), validated_data)


class SerializerMeta:
    model = _Episode
    fields = ["__all__"]
    custom_field_write_permission_map = {"subtitle": {"program.modify_episode_subtitle"}}


class _UpdateFieldPermissionValidatorSerializer(
    UpdateFieldPermissionValidatorMixin, SimpleAssignmentWriter, serializers.Serializer
):
    title = serializers.CharField()
    subtitle = serializers.CharField()
    content = serializers.CharField()

    class Meta(SerializerMeta):
        pass


class TestUpdateFieldPermissionValidatorMixin:
    def test_permissions_are_resolved_according_to_specification(self):
        serializer = _UpdateFieldPermissionValidatorSerializer()
        field_write_permission_map = serializer._get_field_write_permission_map(
            {"title", "subtitle", "content", "image_id", "contributor_ids"}
        )
        assert field_write_permission_map == {
            "title": {"program.edit__episode__title"},
            "subtitle": {"program.modify_episode_subtitle"},
            "image_id": {"program.edit__episode__image"},
            "contributor_ids": {"program.edit__episode__contributors"},
        }

    def test_serializer_raises_if_permissions_are_missing(self):
        instance = _Episode()
        update_data = {
            "title": "my new title",
            "subtitle": "my new subtitle",
            "content": "my new content",
        }
        serializer = _UpdateFieldPermissionValidatorSerializer(
            instance=instance, data=update_data, context=DummyUser.make_context()
        )
        try:
            serializer.is_valid(raise_exception=True)
        except PermissionDenied as exc:
            write_exception = exc
        else:
            assert False, "is_valid should raise"
        details = write_exception.get_full_details()
        assert details["title"]["code"] == "missing-field-permission"
        assert details["subtitle"]["code"] == "missing-field-permission"
        assert "content" not in details

    def test_serializer_succeeds_if_user_has_permissions(self):
        serializer = _UpdateFieldPermissionValidatorSerializer(
            context=DummyUser.make_context(
                ["program.edit__episode__title", "program.modify_episode_subtitle"]
            )
        )
        update_data = {
            "title": "my new title",
            "subtitle": "my new subtitle",
            "content": "my new content",
        }
        instance: _Episode = serializer.update(_Episode(), update_data)
        for key, value in update_data.items():
            assert getattr(instance, key) == value


class TestSimpleAuditUpdaterMixin:
    def test_update_sets_updated_by_field(self):
        class Serializer(SimpleAuditUpdaterMixin, SimpleAssignmentWriter, serializers.Serializer):
            class Meta(SerializerMeta):
                pass

        episode = _Episode()
        serializer = Serializer(context=DummyUser.make_context())
        serializer.update(episode, {})
        assert episode.updated_by == DummyUser().username

    def test_create_sets_created_by_and_updated_by_field(self):
        class Serializer(SimpleAuditUpdaterMixin, SimpleAssignmentWriter, serializers.Serializer):
            class Meta(SerializerMeta):
                pass

        serializer = Serializer(context=DummyUser.make_context())
        username = serializer.context.get("request").user.username
        episode = serializer.create({})
        assert episode.created_by == username
        assert episode.updated_by == username
